const TinyRaft = require('./tinyraft.js');

function newNode(id, nodes, partitions)
{
    partitions = partitions || {};
    let n = new TinyRaft({
        nodes: [ 1, 2, 3, 4, 5 ],
        nodeId: id,
        heartbeatTimeout: 100,
        electionTimeout: 500,
        send: function(to, msg)
        {
            if (!partitions[n.nodeId+'-'+to] && !partitions[to+'-'+n.nodeId] && nodes[to])
            {
                console.log('received from '+n.nodeId+' to '+to+': '+JSON.stringify(msg));
                setImmediate(function() { nodes[to].onReceive(n.nodeId, msg); });
            }
        },
    });
    n.on('change', (st) =>
    {
        console.log(
            'node '+n.nodeId+': '+(st.state == TinyRaft.FOLLOWER ? 'following '+st.leader : st.state)+
            ', term '+st.term+(st.state == TinyRaft.LEADER ? ', followers: '+st.followers.join(', ') : '')
        );
    });
    nodes[id] = n;
}

function newNodes(count, partitions)
{
    partitions = partitions || {};
    const nodes = {};
    for (let i = 1; i <= count; i++)
    {
        newNode(i, nodes, partitions);
    }
    for (let i = 1; i <= count; i++)
    {
        nodes[i].start();
    }
    return nodes;
}

function checkQuorum(nodes, count)
{
    let leaders = 0;
    for (const i in nodes)
    {
        if (nodes[i].state == TinyRaft.LEADER)
        {
            leaders++;
            if (nodes[i].followers.length != count)
                throw new Error('leader '+i+' has '+nodes[i].followers.length+' followers, but should have '+count);
        }
        else if (nodes[i].state != TinyRaft.FOLLOWER)
        {
            throw new Error('node '+i+' is not in quorum: state '+nodes[i].state);
        }
    }
    if (leaders != 1)
    {
        throw new Error('we have '+leaders+' leaders instead of exactly 1');
    }
    console.log('OK: '+count+' nodes in quorum');
}

function checkNoQuorum(nodes)
{
    const nc = Object.values(nodes).filter(n => n.state != TinyRaft.CANDIDATE);
    if (nc.length > 0)
    {
        throw new Error('we have non-candidates ('+nc.map(n => n.nodeId).join(', ')+'), but we should not');
    }
    console.log('OK: '+Object.keys(nodes).length+' candidates, no quorum');
}

async function testStartThenRemoveNode()
{
    console.log('testStartThenRemoveNode');
    const nodes = newNodes(5);
    let leaderChanges = 0, prevLeader = null;
    nodes[2].on('change', (st) =>
    {
        if (st.leader && st.leader != prevLeader)
        {
            prevLeader = st.leader;
            leaderChanges++;
        }
    });
    await new Promise(ok => setTimeout(ok, 2000));
    checkQuorum(nodes, 5);
    if (leaderChanges >= 3)
    {
        throw new Error('More than 3 leader changes in 2000ms: '+leaderChanges);
    }
    // Stop the leader
    let leader = nodes[1].leader;
    nodes[leader].onReceive = () => {};
    nodes[leader].stop();
    delete nodes[leader];
    // Check quorum after 2000ms
    await new Promise(ok => setTimeout(ok, 2000));
    checkQuorum(nodes, 4);
    // Stop the leader again
    leader = nodes[Object.keys(nodes)[0]].leader;
    nodes[leader].onReceive = () => {};
    nodes[leader].stop();
    delete nodes[leader];
    // Check quorum after 2000ms
    await new Promise(ok => setTimeout(ok, 2000));
    checkQuorum(nodes, 3);
    // Stop the leader again
    leader = nodes[Object.keys(nodes)[0]].leader;
    nodes[leader].onReceive = () => {};
    nodes[leader].stop();
    delete nodes[leader];
    // Check that no quorum exists
    await new Promise(ok => setTimeout(ok, 2000));
    checkNoQuorum(nodes);
    // Clean up
    for (const id in nodes)
    {
        nodes[id].stop();
    }
    console.log('testStartThenRemoveNode: OK');
}

async function testAddNode()
{
    console.log('testAddNode');
    const nodes = newNodes(5);
    await new Promise(ok => setTimeout(ok, 2000));
    checkQuorum(nodes, 5);
    // Add node
    newNode(6, nodes);
    for (let i = 1; i <= 5; i++)
        nodes[i].setNodes([ 1, 2, 3, 4, 5, 6 ]);
    nodes[6].start();
    // Check quorum after 2000ms
    await new Promise(ok => setTimeout(ok, 2000));
    checkQuorum(nodes, 6);
    // Clean up
    for (const id in nodes)
    {
        nodes[id].stop();
    }
    console.log('testAddNode: OK');
}

async function run()
{
    await testStartThenRemoveNode();
    await testAddNode();
    process.exit(0);
}

run().catch(err => { console.error(err); process.exit(1); });
